/************************************************************************/
/*
controller for walking (two legs with knees, no feet but foot torques exist)
*/
/***********************************************************************/
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "useful/useful.h"
#include "nr2d/nrutil.h"
#include "dm/dm.h"
#include "trajectory/trajectory.h"
#include "dynamics.h"
#include "controller.h"

/**********************************************************************/
/*****************************************************************************/

Controller *create_controller()
{
  int i;
  Controller *result;

  result = (Controller *) malloc( sizeof( Controller ) );
  if ( result == NULL )
    {
      fprintf( stderr, "Can't allocate dynamics.\n" );
      exit( -1 );
    }

  init_trajectory( result->desired_trajectory, N_TRAJ );
  init_trajectory( result->current_desireds, N_TRAJ );

  result->d = create_dynamics();

  return result;
}

/*****************************************************************************/
/*****************************************************************************/

void load_actuals( Controller *c, Dynamics *d )
{
  int i = 0;

  c->actuals[i] = d->xh;
  c->actualsd[i++] = d->xhd;
  c->actuals[i] = d->yh;
  c->actualsd[i++] = d->yhd;
  c->actuals[i] = d->ato;
  c->actualsd[i++] = d->wto;
  c->actuals[i] = d->xal;
  c->actualsd[i++] = d->xald;
  c->actuals[i] = d->yal;
  c->actualsd[i++] = d->yald;
  c->actuals[i] = d->xar;
  c->actualsd[i++] = d->xard;
  c->actuals[i] = d->yar;
  c->actualsd[i++] = d->yard;
  c->actuals[i] = d->ahl;
  c->actualsd[i++] = d->ahld;
  c->actuals[i] = d->ahr;
  c->actualsd[i++] = d->ahrd;
  c->actuals[i] = d->akl;
  c->actualsd[i++] = d->akld;
  c->actuals[i] = d->akr;
  c->actualsd[i++] = d->akrd;
  c->actuals[i] = d->aal;
  c->actualsd[i++] = d->aald;
  c->actuals[i] = d->aar;
  c->actualsd[i++] = d->aard;
}

/*****************************************************************************/

void plan_launch( Controller *c, double *plan_time,
		  double *xh, double *xal, double *xar )
{

  *xh += c->l1_x;

  add_knot_point( c->desired_trajectory, T_XH, QUINTIC_SPLINE, 
		  *plan_time + c->l1_t1, *xh, c->velocity, 0.0 );

  add_knot_point( c->desired_trajectory, T_YH, QUINTIC_SPLINE, 
		  *plan_time + c->l1_t1, c->height, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_XAL, QUINTIC_SPLINE, 
		  *plan_time + c->l1_tx, *xal, 0.0, 0.0 );

  *xal = c->l1_xal + *xh;

  add_knot_point( c->desired_trajectory, T_XAL, QUINTIC_SPLINE, 
		  *plan_time + c->l1_t1 - c->l1_tx, *xal, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_YAL, QUINTIC_SPLINE, 
		  *plan_time + c->l1_ty, c->clearance, 0.0, 0.0 );
  add_knot_point( c->desired_trajectory, T_YAL, QUINTIC_SPLINE, 
		  *plan_time + c->l1_t1 - c->l1_ty, c->clearance, 0.0, 0.0 );
  add_knot_point( c->desired_trajectory, T_YAL, QUINTIC_SPLINE, 
		  *plan_time + c->l1_t1, 0.0, c->touchdown_speed, 0.0 );
  add_knot_point( c->desired_trajectory, T_YAL, QUINTIC_SPLINE, 
		  *plan_time + c->l1_t1 + c->touchdown_duration,
		  c->touchdown_duration*c->touchdown_speed/2, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_XAR, QUINTIC_SPLINE, 
		  *plan_time, *xar, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_YAR, QUINTIC_SPLINE, 
		  *plan_time, 0.0, 0.0, 0.0 );

  *plan_time += c->l1_t1;
}

/*****************************************************************************/

void plan_ds1( Controller *c, double *plan_time,
	       double *xh, double *xal, double *xar )
{

  *xh += c->velocity*c->ds_time;

  add_knot_point( c->desired_trajectory, T_XH, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, *xh, c->velocity, 0.0 );

  add_knot_point( c->desired_trajectory, T_YH, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, c->height, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_XAL, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, *xal, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_YAL, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, 0.0, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_XAR, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, *xar, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_YAR, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, 0.0, 0.0, 0.0 );

  *plan_time += c->ds_time;
}

/*****************************************************************************/

void plan_ssl( Controller *c, double *plan_time,
	       double *xh, double *xal, double *xar )
{

  *xh += c->velocity*c->ss_time;

  add_knot_point( c->desired_trajectory, T_XH, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time, *xh, c->velocity, 0.0 );

  add_knot_point( c->desired_trajectory, T_YH, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time, c->height, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_XAR, QUINTIC_SPLINE, 
		  *plan_time + c->tx, *xar, 0.0, 0.0 );

  *xar = *xh + c->step_size;

  add_knot_point( c->desired_trajectory, T_XAR, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time - c->tx, *xar, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_YAR, QUINTIC_SPLINE, 
		  *plan_time, 0.0, 0.0, 0.0 );
  add_knot_point( c->desired_trajectory, T_YAR, QUINTIC_SPLINE, 
		  *plan_time + c->ty, c->clearance, 0.0, 0.0 );
  add_knot_point( c->desired_trajectory, T_YAR, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time - c->ty, c->clearance, 0.0, 0.0 );
  add_knot_point( c->desired_trajectory, T_YAR, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time, 0.0, c->touchdown_speed, 0.0 );
  add_knot_point( c->desired_trajectory, T_YAR, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time + c->touchdown_duration,
		  c->touchdown_duration*c->touchdown_speed/2, 0.0, 0.0 );

  *plan_time += c->ss_time;
}

/*****************************************************************************/

void plan_ds2( Controller *c, double *plan_time,
	       double *xh, double *xal, double *xar )
{

  *xh += c->velocity*c->ds_time;

  add_knot_point( c->desired_trajectory, T_XH, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, *xh, c->velocity, 0.0 );

  add_knot_point( c->desired_trajectory, T_YH, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, c->height, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_XAL, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, *xal, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_YAL, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, 0.0, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_XAR, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, *xar, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_YAR, QUINTIC_SPLINE, 
		  *plan_time + c->ds_time, 0.0, 0.0, 0.0 );

  *plan_time += c->ds_time;
}

/*****************************************************************************/

void plan_ssr( Controller *c, double *plan_time,
	       double *xh, double *xal, double *xar )
{

  *xh += c->velocity*c->ss_time;

  add_knot_point( c->desired_trajectory, T_XH, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time, *xh, c->velocity, 0.0 );

  add_knot_point( c->desired_trajectory, T_YH, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time, c->height, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_XAL, QUINTIC_SPLINE, 
		  *plan_time + c->tx, *xal, 0.0, 0.0 );

  *xal = *xh + c->step_size;

  add_knot_point( c->desired_trajectory, T_XAL, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time - c->tx, *xal, 0.0, 0.0 );

  add_knot_point( c->desired_trajectory, T_YAL, QUINTIC_SPLINE, 
		  *plan_time, 0.0, 0.0, 0.0 );
  add_knot_point( c->desired_trajectory, T_YAL, QUINTIC_SPLINE, 
		  *plan_time + c->ty, c->clearance, 0.0, 0.0 );
  add_knot_point( c->desired_trajectory, T_YAL, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time - c->ty, c->clearance, 0.0, 0.0 );
  add_knot_point( c->desired_trajectory, T_YAL, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time, 0.0, c->touchdown_speed, 0.0 );
  add_knot_point( c->desired_trajectory, T_YAL, QUINTIC_SPLINE, 
		  *plan_time + c->ss_time + c->touchdown_duration,
		  c->touchdown_duration*c->touchdown_speed/2, 0.0, 0.0 );

  *plan_time += c->ss_time;
}

/*****************************************************************************/

void init_controller( Controller *c, Dynamics *d )
{
  int i;
  double plan_time, xh, xal, xar;

  c->start_time = d->time;
  c->max_height = d->shank_length + d->thigh_length;

  /*
  c->d->dynamics_state = D_DS;
  c->controller_state = C_STANDING;
  */

  c->d->dynamics_state = D_SSR;
  c->controller_state = C_LAUNCH;

  c->height = c->max_height - 0.1;
  c->velocity = 1.0;
  c->step_size = 0.4;
  c->ds_time = 0.1;
  c->ss_time = 0.4;
  c->tx = c->ss_time*0.1;
  c->ty = c->ss_time*0.3;
  c->clearance = 0.03;
  c->touchdown_speed = -0.1;
  c->touchdown_duration = 0.02;

  c->l1_t1 = 0.4;
  c->l1_tx = c->l1_t1*0.1;
  c->l1_ty = c->l1_t1*0.3;
  c->l1_x = 0.2;
  c->l1_xal = c->step_size;

  c->k_p[T_AHL] = 3000.0;
  c->k_p[T_AHR] = 3000.0;
  c->k_p[T_AKL] = -3000.0;
  c->k_p[T_AKR] = -3000.0;
  c->k_p[T_AAL] = -3000.0;
  c->k_p[T_AAR] = -3000.0;

  c->k_v[T_AHL] = 300.0;
  c->k_v[T_AHR] = 300.0;
  c->k_v[T_AKL] = -300.0;
  c->k_v[T_AKR] = -300.0;
  c->k_v[T_AAL] = -300.0;
  c->k_v[T_AAR] = -300.0;

  // get useful variables in vector form
  load_actuals( c, d );

  // add desired position as knot to current desireds
  for ( i = 0; i < N_TRAJ; i++ )
    {
      add_knot_point( c->current_desireds, i, QUINTIC_SPLINE, 0.0,
		      c->actuals[i], 0.0, 0.0 );
      // printf( "%d %g\n", i, c->actuals[i] );
    }

  // initialize some variables
  for ( i = 0; i < N_ACC; i++ )
    c->qdd_d[i] = 0;
  for ( i = 0; i < N_DS_F; i++ )
    c->forces_d[i] = 0;
  c->forces_d[DS_F_FGSLY] = 441.45;
  c->forces_d[DS_F_FGSRY] = 441.45;

  // make a default plan

  plan_time = d->time;
  xh = d->xh;
  xal = d->xal;
  xar = d->xar;

  // For walking, create an initial plan
  plan_launch( c, &plan_time, &xh, &xal, &xar ); // plan two phases ahead
  plan_ds1( c, &plan_time, &xh, &xal, &xar );
}

/*****************************************************************************/
/*****************************************************************************/

void run_trajectories( Controller *c, double time )
{
  int i;
  double position, velocity, acceleration;

  for ( i = 0; i < N_TRAJ; i++ )
    {
      if ( !lookup_spline3( c->desired_trajectory, i, time,
			    c->current_desireds, &position,
			    &velocity, &acceleration ) )
	{
	  printf( "Lookup failure.\n" );
	  exit( -1 );
	}
      /* update current desired position */
      set_first_knot_to( c->current_desireds, i, QUINTIC_SPLINE, time,
			 position, velocity, acceleration );
      c->desireds[i] = position;
      c->desiredsd[i] = velocity;
      c->desiredsdd[i] = acceleration;
      /*
      printf( "%d %g %g %g\n", i, c->desireds[i], c->desiredsd[i],
	      c->desiredsdd[i] );
      */
    }
}

/*****************************************************************************/
/*****************************************************************************/
// actions for each state.

int do_launch( Controller *c, Dynamics *d )
{
  // nothing to do.
}

/*****************************************************************************/

int do_ds1( Controller *c, Dynamics *d )
{
  // nothing to do.
}

/*****************************************************************************/

int do_ssl( Controller *c, Dynamics *d )
{
  // nothing to do.
}

/*****************************************************************************/

int do_ds2( Controller *c, Dynamics *d )
{
  // nothing to do.
}

/*****************************************************************************/

int do_ssr( Controller *c, Dynamics *d )
{
  // nothing to do.
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

int handle_launch( Controller *c, Dynamics *d )
{
  double plan_time, xh, xal, xar;

  // check for transition to double support (C_DS1)
  if ( d->time - c->start_time > c->l1_t1 - 0.1 && d->s_fy_l > 1.0 )
    {
      c->controller_state = C_DS1;
      c->d->dynamics_state = D_DS; // use double support inverse dynamics
      c->start_time = d->time;
      plan_time = d->time;
      xh = d->xh;
      xal = d->xal;
      xar = d->xar;
      clear_all_knots( c->desired_trajectory, N_TRAJ );
      plan_ds1( c, &plan_time, &xh, &xal, &xar ); // plan 2 phases ahead
      plan_ssl( c, &plan_time, &xh, &xal, &xar );
      run_trajectories( c, d->time ); // redo this with new knot points
      do_ds1( c, d );
      return 0; // transition occurred
    }

  do_launch( c, d );
  return 1; // no transition
}


/*****************************************************************************/

int handle_ds1( Controller *c, Dynamics *d )
{
  double plan_time, xh, xal, xar;

  // check for transition to single support (C_SSL)
  if ( d->time - c->start_time > c->ds_time ) // -0.01?
    {
      c->controller_state = C_SSL;
      c->d->dynamics_state = D_SSL; // use double support inverse dynamics
      c->start_time = d->time;
      plan_time = d->time;
      xh = d->xh;
      xal = d->xal;
      xar = d->xar;
      clear_all_knots( c->desired_trajectory, N_TRAJ );
      plan_ssl( c, &plan_time, &xh, &xal, &xar ); // plan 2 phases ahead
      plan_ds2( c, &plan_time, &xh, &xal, &xar );
      run_trajectories( c, d->time ); // redo this with new knot points
      do_ssl( c, d );
      return 0; // transition occurred
    }

  do_ds1( c, d );
  return 1; // no transition
}

/*****************************************************************************/

int handle_ssl( Controller *c, Dynamics *d )
{
  double plan_time, xh, xal, xar;

  // check for transition to double support (C_DS2)
  if ( d->time - c->start_time > c->ss_time - 0.1 && d->s_fy_r > 1.0 )
    {
      c->controller_state = C_DS2;
      c->d->dynamics_state = D_DS; // use double support inverse dynamics
      c->start_time = d->time;
      plan_time = d->time;
      xh = d->xh;
      xal = d->xal;
      xar = d->xar;
      clear_all_knots( c->desired_trajectory, N_TRAJ );
      plan_ds2( c, &plan_time, &xh, &xal, &xar ); // plan 2 phases ahead
      plan_ssr( c, &plan_time, &xh, &xal, &xar );
      run_trajectories( c, d->time ); // redo this with new knot points
      do_ds2( c, d );
      return 0; // transition occurred
    }

  do_ssl( c, d );
  return 1; // no transition
}

/*****************************************************************************/

int handle_ds2( Controller *c, Dynamics *d )
{
  double plan_time, xh, xal, xar;

  // check for transition to single support (C_SSR)
  if ( d->time - c->start_time > c->ds_time ) // -0.01?
    {
      c->controller_state = C_SSR;
      c->d->dynamics_state = D_SSR; // use double support inverse dynamics
      c->start_time = d->time;
      plan_time = d->time;
      xh = d->xh;
      xal = d->xal;
      xar = d->xar;
      clear_all_knots( c->desired_trajectory, N_TRAJ );
      plan_ssr( c, &plan_time, &xh, &xal, &xar ); // plan 2 phases ahead
      plan_ds1( c, &plan_time, &xh, &xal, &xar );
      run_trajectories( c, d->time ); // redo this with new knot points
      do_ssr( c, d );
      return 0; // transition occurred
    }

  do_ds2( c, d );
  return 1; // no transition
}

/*****************************************************************************/

int handle_ssr( Controller *c, Dynamics *d )
{
  double plan_time, xh, xal, xar;

  // check for transition to double support (C_DS1)
  if ( d->time - c->start_time > c->ss_time - 0.1 && d->s_fy_l > 1.0 )
    {
      c->controller_state = C_DS1;
      c->d->dynamics_state = D_DS; // use double support inverse dynamics
      c->start_time = d->time;
      plan_time = d->time;
      xh = d->xh;
      xal = d->xal;
      xar = d->xar;
      clear_all_knots( c->desired_trajectory, N_TRAJ );
      plan_ds1( c, &plan_time, &xh, &xal, &xar ); // plan 2 phases ahead
      plan_ssl( c, &plan_time, &xh, &xal, &xar );
      run_trajectories( c, d->time ); // redo this with new knot points
      do_ds1( c, d );
      return 0; // transition occurred
    }

  do_ssr( c, d );
  return 1; // no transition
}

/*****************************************************************************/

void do_inverse_dynamics( Controller *c, Dynamics *d )
{

  c->d->time = 0;

  c->d->xh = c->desireds[T_XH];
  c->d->yh = c->desireds[T_YH];
  c->d->ato = c->desireds[T_ATO];
  c->d->xal = c->desireds[T_XAL];
  c->d->yal = c->desireds[T_YAL];
  c->d->xar = c->desireds[T_XAR];
  c->d->yar = c->desireds[T_YAR];

  c->d->xhd = c->desiredsd[T_XH];
  c->d->yhd = c->desiredsd[T_YH];
  c->d->wto = c->desiredsd[T_ATO];
  c->d->xald = c->desiredsd[T_XAL];
  c->d->yald = c->desiredsd[T_YAL];
  c->d->xard = c->desiredsd[T_XAR];
  c->d->yard = c->desiredsd[T_YAR];

  c->d->xhdd = c->desiredsdd[T_XH];
  c->d->yhdd = c->desiredsdd[T_YH];
  c->d->wtod = c->desiredsdd[T_ATO];
  c->d->xaldd = c->desiredsdd[T_XAL];
  c->d->yaldd = c->desiredsdd[T_YAL];
  c->d->xardd = c->desiredsdd[T_XAR];
  c->d->yardd = c->desiredsdd[T_YAR];

  set_state( c->d );

  c->desireds[T_AHL] = c->d->full_state[X_AHL];
  c->desireds[T_AHR] = c->d->full_state[X_AHR];
  c->desireds[T_AKL] = c->d->full_state[X_AKL];
  c->desireds[T_AKR] = c->d->full_state[X_AKR];
  c->desireds[T_AAL] = c->d->asl;
  c->desireds[T_AAR] = c->d->asr;

  c->desiredsd[T_AHL] = c->d->full_state[X_AHLD];
  c->desiredsd[T_AHR] = c->d->full_state[X_AHRD];
  c->desiredsd[T_AKL] = c->d->full_state[X_AKLD];
  c->desiredsd[T_AKR] = c->d->full_state[X_AKRD];
  c->desiredsd[T_AAL] = c->d->wsl;
  c->desiredsd[T_AAR] = c->d->wsr;

  c->desiredsdd[T_AHL] = c->d->accelerations[X_AHL];
  c->desiredsdd[T_AHR] = c->d->accelerations[X_AHR];
  c->desiredsdd[T_AKL] = c->d->accelerations[X_AKL];
  c->desiredsdd[T_AKR] = c->d->accelerations[X_AKR];
  c->desiredsdd[T_AAL] = c->d->accelerations[X_ATO]
    + c->d->accelerations[X_AHL] - c->d->accelerations[X_AKL]; 
  c->desiredsdd[T_AAR] = c->d->accelerations[X_ATO]
    + c->d->accelerations[X_AHR] - c->d->accelerations[X_AKR];

  c->qdd_d[X_XH] = c->desiredsdd[T_XH];
  c->qdd_d[X_YH] = c->desiredsdd[T_YH];
  c->qdd_d[X_ATO] = c->desiredsdd[T_ATO];
  c->qdd_d[X_AHL] = c->d->accelerations[X_AHL];
  c->qdd_d[X_AHR] = c->d->accelerations[X_AHR];
  c->qdd_d[X_AKL] = c->d->accelerations[X_AKL];
  c->qdd_d[X_AKR] = c->d->accelerations[X_AKR];

  inverse_dynamics( c->d, c->qdd_d, c->forces_d );

  /*
  printf( "final forces: " );
  for ( j = 0; j < N_DS_F; j++ )
    printf( " %g", c->forces_d[j] );
  printf( "\n" );
  */
}

/*****************************************************************************/

void do_feedback_control( Controller *c, Dynamics *d )
{

  c->forces_d[U_AHL] -= c->k_p[T_AHL]*(c->actuals[T_AHL] - c->desireds[T_AHL])
    + c->k_v[T_AHL]*(c->actualsd[T_AHL] - c->desiredsd[T_AHL]);
  c->forces_d[U_AHR] -= c->k_p[T_AHR]*(c->actuals[T_AHR] - c->desireds[T_AHR])
    + c->k_v[T_AHR]*(c->actualsd[T_AHR] - c->desiredsd[T_AHR]);
  c->forces_d[U_AKL] -= c->k_p[T_AKL]*(c->actuals[T_AKL] - c->desireds[T_AKL])
    + c->k_v[T_AKL]*(c->actualsd[T_AKL] - c->desiredsd[T_AKL]);
  c->forces_d[U_AKR] -= c->k_p[T_AKR]*(c->actuals[T_AKR] - c->desireds[T_AKR])
    + c->k_v[T_AKR]*(c->actualsd[T_AKR] - c->desiredsd[T_AKR]);

  // Don't torque foot if it is not on the ground
  if ( d->s_fy_l > 1.0 )
    c->forces_d[U_AAL] -=
      c->k_p[T_AAL]*(c->actuals[T_AAL] - c->desireds[T_AAL])
      + c->k_v[T_AAL]*(c->actualsd[T_AAL] - c->desiredsd[T_AAL]);
  else
    c->forces_d[U_AAL] = 0.0;

  // Don't torque foot if it is not on the ground
  if ( d->s_fy_r > 1.0 )
    c->forces_d[U_AAR] -=
      c->k_p[T_AAR]*(c->actuals[T_AAR] - c->desireds[T_AAR])
      + c->k_v[T_AAR]*(c->actualsd[T_AAR] - c->desiredsd[T_AAR]);
  else
    c->forces_d[U_AAR] = 0.0;
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

run_controller( Controller *c, Dynamics *d )
{
  int i, j;
 
  // printf( "\nrun_controller: %g %d\n", d->time, d->dynamics_state );

  // get useful variables in vector form
  load_actuals( c, d );

  // use the trajectories.
  run_trajectories( c, d->time );

  // do controller state dependent stuff
  switch( c->controller_state )
    {
    case C_STANDING:
      // do nothing
      break;
    case C_LAUNCH:
      handle_launch( c, d );
      break;
    case C_DS1:
      handle_ds1( c, d );
      break;
    case C_SSL:
      handle_ssl( c, d );
      break;
    case C_DS2:
      handle_ds2( c, d );
      break;
    case C_SSR:
      handle_ssr( c, d );
      break;
    default:
      fprintf( stderr, "Unrecognized controller state %d in run_controller\n",
	       c->controller_state );
      exit( -1 );
    }

  do_inverse_dynamics( c, d );

  do_feedback_control( c, d );

  // copy desired forces into dynamics structure
  for ( i = 0; i < N_U; i++ )
    d->action[i] = c->forces_d[i];
}

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
