/**********************************************************************/

/* Controller states */
#define C_UNKNOWN 0
#define C_STANDING 1
#define C_DS1 2  // double support, left leg in front
#define C_SSL 3 // first phase of single support, left leg in stance
#define C_DS2 4  // double support, right leg in front
#define C_SSR 5 // first phase of single support, right leg in stance
#define C_LAUNCH 6 // launch state. Like DS1. largest value so plot is pretty
#define N_CSTATES 7

/*****************************************************************************/
/* Trajectory variables */

#define T_XH 0 // Cartesian variables
#define T_YH 1
#define T_ATO 2
#define T_XAL 3
#define T_YAL 4
#define T_XAR 5
#define T_YAR 6
#define T_AHL 7 // Joint variables
#define T_AHR 8
#define T_AKL 9
#define T_AKR 10
#define T_AAL 11
#define T_AAR 12
#define N_TRAJ 13

/*****************************************************************************/

/* Treat the contents of this structure as hidden */
typedef struct controller_struct
{

  int controller_state;
  double start_time; // what time was the last controller state transition?
  DOFS desired_trajectory[N_TRAJ];
  DOFS current_desireds[N_TRAJ];
  double desireds[N_TRAJ]; // desired positions for servos
  double desiredsd[N_TRAJ]; // desired velocities for servos
  double desiredsdd[N_TRAJ]; // desired accelerations for servos
  double actuals[N_TRAJ]; // actual position values for servos
  double actualsd[N_TRAJ]; // actual velocity values for servos
  double k_p[N_TRAJ]; // position gains
  double k_v[N_TRAJ]; // velocity gains
  double qdd_d[N_ACC]; // desired accelerations
  double forces_d[N_DS_F]; // current best guess as to desired forces
  Dynamics *d; // controllers copy of dynamics.

  // pattern variables
  double max_height;
  double height;
  double velocity;
  double step_size;
  double ss_time;
  double ds_time;
  double clearance;
  double tx;
  double ty;
  double touchdown_speed;
  double touchdown_duration;
  double l1_t1;
  double l1_tx;
  double l1_ty;
  double l1_x;
  double l1_xal;

} Controller;

/*****************************************************************************/

Controller *create_controller();

/*****************************************************************************/

