/*****************************************************************************/
/*
  data.c: save data from simulation
*/
/*****************************************************************************/
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mrdplot/mrdplot.h"

#include "useful/useful.h"
#include "trajectory/trajectory.h"
#include "dynamics.h"
#include "controller.h"

#include "data.h"

/*****************************************************************************/

/*****************************************************************************/

int n_points = 0;
float data[MAX_N_POINTS*N_CHANNELS];

char *names[N_CHANNELS];
char *units[N_CHANNELS];

/*****************************************************************************/

void init_data( Dynamics *s )
{
  int i;

  /*
  printf( "N_CHANNELS: %d; M_L_TORSO_BOTTOM: %d; last: %d\n", N_CHANNELS,
	  M_L_TORSO_BOTTOM, M_R_ANKLE_COMMAND );
  */

  for( i = 0; i < N_CHANNELS; i++ )
    {
      names[i] = "xxx";
      units[i] = "xxx";
    }

  names[M_TIME] = "time";

  names[M_DYNAMICS_STATE] = "d_state";
  names[M_CONTROLLER_STATE] = "c_state";
  names[M_CONTROLLER_DYNAMICS_STATE] = "c_d_state";

  names[M_X_XH] = "x_xh";
  names[M_X_YH] = "x_yh";
  names[M_X_ATO] = "x_ato";
  names[M_X_AHL] = "x_ahl";
  names[M_X_AHR] = "x_ahr"; 
  names[M_X_AKL] = "x_akl";
  names[M_X_AKR] = "x_akr"; 
  names[M_ASL] = "asl";
  names[M_ASR] = "asr";

  names[M_XH_D] = "xh_d";
  names[M_YH_D] = "yh_d";
  names[M_ATO_D] = "ato_d";
  names[M_AHL_D] = "ahl_d";
  names[M_AHR_D] = "ahr_d"; 
  names[M_AKL_D] = "akl_d";
  names[M_AKR_D] = "akr_d"; 
  names[M_AAL_D] = "aal_d";
  names[M_AAR_D] = "aar_d";

  names[M_X_XHD] = "x_xhd";
  names[M_X_YHD] = "x_yhd";
  names[M_X_ATOD] = "x_atod";
  names[M_X_AHLD] = "x_ahld";
  names[M_X_AHRD] = "x_ahrd"; 
  names[M_X_AKLD] = "x_akld";
  names[M_X_AKRD] = "x_akrd"; 
  names[M_WSL] = "wsl";
  names[M_WSR] = "wsr";

  names[M_XHD_D] = "xhd_d";
  names[M_YHD_D] = "yhd_d";
  names[M_ATOD_D] = "atod_d";
  names[M_AHLD_D] = "ahld_d";
  names[M_AHRD_D] = "ahrd_d"; 
  names[M_AKLD_D] = "akld_d";
  names[M_AKRD_D] = "akrd_d"; 
  names[M_AALD_D] = "aal_d";
  names[M_AARD_D] = "aar_d";

  names[M_U_AHL] = "u_ahl";
  names[M_U_AKL] = "u_akl";
  names[M_U_AAL] = "u_aal";
  names[M_U_AHR] = "u_ahr";
  names[M_U_AKR] = "u_akr";
  names[M_U_AAR] = "u_aar";

  names[M_U_AHL_D] = "u_ahl_d";
  names[M_U_AKL_D] = "u_akl_d";
  names[M_U_AAL_D] = "u_aal_d";
  names[M_U_AHR_D] = "u_ahr_d";
  names[M_U_AKR_D] = "u_akr_d";
  names[M_U_AAR_D] = "u_aar_d";

  names[M_FGSLX] = "FGSLx";
  names[M_FGSLY] = "FGSLy";
  names[M_FGSRX] = "FGSRx";
  names[M_FGSRY] = "FGSRy";

  names[M_XAL] = "xal";
  names[M_YAL] = "yal";
  names[M_XAR] = "xar";
  names[M_YAR] = "yar";

  names[M_XAL_D] = "xal_d";
  names[M_YAL_D] = "yal_d";
  names[M_XAR_D] = "xar_d";
  names[M_YAR_D] = "yar_d";

  names[M_XALD] = "xald";
  names[M_YALD] = "yald";
  names[M_XARD] = "xard";
  names[M_YARD] = "yard";

  names[M_XALD_D] = "xald_d";
  names[M_YALD_D] = "yald_d";
  names[M_XARD_D] = "xard_d";
  names[M_YARD_D] = "yard_d";

  names[M_X_XHDD] = "x_xhdd";
  names[M_X_YHDD] = "x_yhdd";
  names[M_X_ATODD] = "x_atodd";
  names[M_X_AHLDD] = "x_ahldd";
  names[M_X_AHRDD] = "x_ahrdd"; 
  names[M_X_AKLDD] = "x_akldd";
  names[M_X_AKRDD] = "x_akrdd"; 
  names[M_WSLD] = "wsld";
  names[M_WSRD] = "wsrd";

  names[M_XHDD_D] = "xhdd_d";
  names[M_YHDD_D] = "yhdd_d";
  names[M_ATODD_D] = "atodd_d";
  names[M_AHLDD_D] = "ahldd_d";
  names[M_AHRDD_D] = "ahrdd_d"; 
  names[M_AKLDD_D] = "akldd_d";
  names[M_AKRDD_D] = "akrdd_d"; 
  names[M_AALDD_D] = "aaldd_d";
  names[M_AARDD_D] = "aardd_d"; 

  names[M_XSL] = "xsl";
  names[M_YSL] = "ysl";
  names[M_XKL] = "xkl";
  names[M_YKL] = "ykl";
  names[M_XTHL] = "xthl";
  names[M_YTHL] = "ythl"; 
  names[M_ATHL] = "athl"; 
  names[M_XSR] = "xsr";
  names[M_YSR] = "ysr";
  names[M_XKR] = "xkr";
  names[M_YKR] = "ykr";
  names[M_XTHR] = "xthr";
  names[M_YTHR] = "ythr"; 
  names[M_ATHR] = "athr"; 
  names[M_XH] = "xh";
  names[M_YH] = "yh";
  names[M_XTO] = "xto";
  names[M_YTO] = "yto";
  names[M_ATO] = "ato";
  names[M_XHEAD] = "xhead";
  names[M_YHEAD] = "yhead";

  names[M_XSLD] = "xsld";
  names[M_YSLD] = "ysld";
  names[M_XKLD] = "xkld";
  names[M_YKLD] = "ykld";
  names[M_XTHLD] = "xthld";
  names[M_YTHLD] = "ythld";
  names[M_XSRD] = "xsrd";
  names[M_YSRD] = "ysrd";
  names[M_XKRD] = "xkrd";
  names[M_YKRD] = "ykrd";
  names[M_XTHRD] = "xthrd";
  names[M_YTHRD] = "ythrd";
  names[M_XHD] = "xhd";
  names[M_YHD] = "yhd";
  names[M_XTOD] = "xtod";
  names[M_YTOD] = "ytod";
  names[M_XHEADD] = "xheadd";
  names[M_YHEADD] = "yheadd";

  names[M_WTHL] = "wthl";
  names[M_WTHR] = "wthr";
  names[M_WTO] = "wto";

  names[M_XCOM] = "xcom";
  names[M_YCOM] = "ycom";

  names[M_XCOMD] = "xcomd";
  names[M_YCOMD] = "ycomd";

  names[M_ENERGY] = "energy";
  names[M_LNE_ENERGY] = "lne_energy";

  names[M_SLKE] = "slke";
  names[M_SLPE] = "slpe";
  names[M_THLKE] = "thlke";
  names[M_THLPE] = "thlpe";
  names[M_SRKE] = "srke";
  names[M_SRPE] = "srpe";
  names[M_THRKE] = "thrke";
  names[M_THRPE] = "thrpe";
  names[M_TOKE] = "toke";
  names[M_TOPE] = "tope";
  names[M_KE] = "ke";
  names[M_PE] = "pe";

  units[M_TIME] = "s";

  units[M_U_AAL] = "N/m";
  units[M_U_AKL] = "N/m";
  units[M_U_AHL] = "N/m";
  units[M_U_AAR] = "N/m";
  units[M_U_AKR] = "N/m";
  units[M_U_AHR] = "N/m";

  units[M_XSL] = "m";
  units[M_YSL] = "m";
  units[M_XKL] = "m";
  units[M_YKL] = "m";
  units[M_XTHL] = "m";
  units[M_YTHL] = "m";
  units[M_XH] = "m";
  units[M_YH] = "m";
  units[M_XTO] = "m";
  units[M_YTO] = "m";
  units[M_XHEAD] = "m";
  units[M_YHEAD] = "m";

  units[M_XSLD] = "m/s";
  units[M_YSLD] = "m/s";
  units[M_XKLD] = "m/s";
  units[M_YKLD] = "m/s";
  units[M_XTHLD] = "m/s";
  units[M_YTHLD] = "m/s";
  units[M_XHD] = "m/s";
  units[M_YHD] = "m/s";
  units[M_XTOD] = "m/s";
  units[M_YTOD] = "m/s";
  units[M_XHEADD] = "m/s";
  units[M_YHEADD] = "m/s";

  units[M_WSL] = "r/s";
  units[M_WTHL] = "r/s";
  units[M_WSR] = "r/s";
  units[M_WTHR] = "r/s";
  units[M_WTO] = "r/s";

  units[M_XCOM] = "m";
  units[M_YCOM] = "m";

  units[M_XCOMD] = "m/s";
  units[M_YCOMD] = "m/s";

  units[M_ENERGY] = "-";
  units[M_LNE_ENERGY] = "-";

  units[M_SLKE] = "-";
  units[M_SLPE] = "-";
  units[M_THLKE] = "-";
  units[M_THLPE] = "-";
  units[M_TOKE] = "-";
  units[M_TOPE] = "-";
  units[M_KE] = "-";
  units[M_PE] = "-";

  /*
  printf( "%d channels, %d MAX_N_POINTS\n", N_CHANNELS, MAX_N_POINTS );

  for( i = 0; i < N_CHANNELS; i++ )
    {
      printf( "%d %s %s\n", i, names[i], units[i] );
    }
  */
}

/*****************************************************************************/

int save_data( Dynamics *d, Controller *c )
{
  static int myindex = 0;
  
  if ( n_points >= MAX_N_POINTS )
    return 0;

  data[myindex + M_TIME] = (float) (d->time);

  data[myindex + M_DYNAMICS_STATE] = (float) (d->dynamics_state);
  data[myindex + M_CONTROLLER_STATE] = (float) (c->controller_state);
  data[myindex + M_CONTROLLER_DYNAMICS_STATE] = (float) (c->d->dynamics_state);

  data[myindex + M_X_XH] = (float) (d->full_state[X_XH]);
  data[myindex + M_X_YH] = (float) (d->full_state[X_YH]);
  data[myindex + M_X_ATO] = (float) (d->full_state[X_ATO]);
  data[myindex + M_X_AHL] = (float) (d->full_state[X_AHL]);
  data[myindex + M_X_AHR] = (float) (d->full_state[X_AHR]);
  data[myindex + M_X_AKL] = (float) (d->full_state[X_AKL]);
  data[myindex + M_X_AKR] = (float) (d->full_state[X_AKR]);

  data[myindex + M_XH_D] = (float) (c->desireds[T_XH]);
  data[myindex + M_YH_D] = (float) (c->desireds[T_YH]);
  data[myindex + M_ATO_D] = (float) (c->desireds[T_ATO]);
  data[myindex + M_AHL_D] = (float) (c->desireds[T_AHL]);
  data[myindex + M_AHR_D] = (float) (c->desireds[T_AHR]);
  data[myindex + M_AKL_D] = (float) (c->desireds[T_AKL]);
  data[myindex + M_AKR_D] = (float) (c->desireds[T_AKR]);
  data[myindex + M_AAL_D] = (float) (c->desireds[T_AAL]);
  data[myindex + M_AAR_D] = (float) (c->desireds[T_AAR]);

  data[myindex + M_X_XHD] = (float) (d->full_state[X_XHD]);
  data[myindex + M_X_YHD] = (float) (d->full_state[X_YHD]);
  data[myindex + M_X_ATOD] = (float) (d->full_state[X_ATOD]);
  data[myindex + M_X_AHLD] = (float) (d->full_state[X_AHLD]);
  data[myindex + M_X_AHRD] = (float) (d->full_state[X_AHRD]);
  data[myindex + M_X_AKLD] = (float) (d->full_state[X_AKLD]);
  data[myindex + M_X_AKRD] = (float) (d->full_state[X_AKRD]);

  data[myindex + M_XHD_D] = (float) (c->desiredsd[T_XH]);
  data[myindex + M_YHD_D] = (float) (c->desiredsd[T_YH]);
  data[myindex + M_ATOD_D] = (float) (c->desiredsd[T_ATO]);
  data[myindex + M_AHLD_D] = (float) (c->desiredsd[T_AHL]);
  data[myindex + M_AHRD_D] = (float) (c->desiredsd[T_AHR]);
  data[myindex + M_AKLD_D] = (float) (c->desiredsd[T_AKL]);
  data[myindex + M_AKRD_D] = (float) (c->desiredsd[T_AKR]);
  data[myindex + M_AALD_D] = (float) (c->desiredsd[T_AAL]);
  data[myindex + M_AARD_D] = (float) (c->desiredsd[T_AAR]);

  data[myindex + M_U_AHL] = (float) (d->action[U_AHL]);
  data[myindex + M_U_AKL] = (float) (d->action[U_AKL]);
  data[myindex + M_U_AAL] = (float) (d->action[U_AAL]);
  data[myindex + M_U_AHR] = (float) (d->action[U_AHR]);
  data[myindex + M_U_AKR] = (float) (d->action[U_AKR]);
  data[myindex + M_U_AAR] = (float) (d->action[U_AAR]);

  data[myindex + M_U_AHL_D] = (float) (c->forces_d[U_AHL]);
  data[myindex + M_U_AKL_D] = (float) (c->forces_d[U_AKL]);
  data[myindex + M_U_AAL_D] = (float) (c->forces_d[U_AAL]);
  data[myindex + M_U_AHR_D] = (float) (c->forces_d[U_AHR]);
  data[myindex + M_U_AKR_D] = (float) (c->forces_d[U_AKR]);
  data[myindex + M_U_AAR_D] = (float) (c->forces_d[U_AAR]);

  /*
  data[myindex + M_FGSLX] = (float) (d->nr_b[DS_B_FGSLX+1]);
  data[myindex + M_FGSLY] = (float) (d->nr_b[DS_B_FGSLY+1]);
  data[myindex + M_FGSRX] = (float) (d->nr_b[DS_B_FGSRX+1]);
  data[myindex + M_FGSRY] = (float) (d->nr_b[DS_B_FGSRY+1]);
  */
  data[myindex + M_FGSLX] = (float) (d->s_fx_l);
  data[myindex + M_FGSLY] = (float) (d->s_fy_l);
  data[myindex + M_FGSRX] = (float) (d->s_fx_r);
  data[myindex + M_FGSRY] = (float) (d->s_fy_r);

  data[myindex + M_XAL] = (float) (d->xal);
  data[myindex + M_YAL] = (float) (d->yal);
  data[myindex + M_XAR] = (float) (d->xar);
  data[myindex + M_YAR] = (float) (d->yar);

  data[myindex + M_XAL_D] = (float) (c->desireds[T_XAL]);
  data[myindex + M_YAL_D] = (float) (c->desireds[T_YAL]);
  data[myindex + M_XAR_D] = (float) (c->desireds[T_XAR]);
  data[myindex + M_YAR_D] = (float) (c->desireds[T_YAR]);

  data[myindex + M_XALD] = (float) (d->xald);
  data[myindex + M_YALD] = (float) (d->yald);
  data[myindex + M_XARD] = (float) (d->xard);
  data[myindex + M_YARD] = (float) (d->yard);

  data[myindex + M_XALD_D] = (float) (c->desiredsd[T_XAL]);
  data[myindex + M_YALD_D] = (float) (c->desiredsd[T_YAL]);
  data[myindex + M_XARD_D] = (float) (c->desiredsd[T_XAR]);
  data[myindex + M_YARD_D] = (float) (c->desiredsd[T_YAR]);

  data[myindex + M_X_XHDD] = (float) (d->accelerations[X_XH]); 
  data[myindex + M_X_YHDD] = (float) (d->accelerations[X_YH]); 
  data[myindex + M_X_ATODD] = (float) (d->accelerations[X_ATO]); 
  data[myindex + M_X_AHLDD] = (float) (d->accelerations[X_AHL]); 
  data[myindex + M_X_AHRDD] = (float) (d->accelerations[X_AHR]); 
  data[myindex + M_X_AKLDD] = (float) (d->accelerations[X_AKL]); 
  data[myindex + M_X_AKRDD] = (float) (d->accelerations[X_AKR]); 
  data[myindex + M_WSLD] = (float) (d->accelerations[X_ATO]
				     + d->accelerations[X_AHL]
				     - d->accelerations[X_AKL]); 
  data[myindex + M_WSRD] = (float) (d->accelerations[X_ATO]
				     + d->accelerations[X_AHR]
				     - d->accelerations[X_AKR]); 

  data[myindex + M_XHDD_D] = (float) (c->desiredsdd[T_XH]);
  data[myindex + M_YHDD_D] = (float) (c->desiredsdd[T_YH]);
  data[myindex + M_ATODD_D] = (float) (c->desiredsdd[T_ATO]); 
  data[myindex + M_AHLDD_D] = (float) (c->desiredsdd[T_AHL]); 
  data[myindex + M_AHRDD_D] = (float) (c->desiredsdd[T_AHR]); 
  data[myindex + M_AKLDD_D] = (float) (c->desiredsdd[T_AKL]); 
  data[myindex + M_AKRDD_D] = (float) (c->desiredsdd[T_AKR]); 
  data[myindex + M_AALDD_D] = (float) (c->desiredsdd[T_AAL]); 
  data[myindex + M_AARDD_D] = (float) (c->desiredsdd[T_AAR]); 

  data[myindex + M_XSL] = (float) (d->xsl);
  data[myindex + M_YSL] = (float) (d->ysl);
  data[myindex + M_ASL] = (float) (d->asl);
  data[myindex + M_XKL] = (float) (d->xkl);
  data[myindex + M_YKL] = (float) (d->ykl);
  data[myindex + M_XTHL] = (float) (d->xthl);
  data[myindex + M_YTHL] = (float) (d->ythl);
  data[myindex + M_ATHL] = (float) (d->athl);
  data[myindex + M_XSR] = (float) (d->xsr);
  data[myindex + M_YSR] = (float) (d->ysr);
  data[myindex + M_ASR] = (float) (d->asr);
  data[myindex + M_XKR] = (float) (d->xkr);
  data[myindex + M_YKR] = (float) (d->ykr);
  data[myindex + M_XTHR] = (float) (d->xthr);
  data[myindex + M_YTHR] = (float) (d->ythr);
  data[myindex + M_ATHR] = (float) (d->athr);
  data[myindex + M_XH] = (float) (d->xh);
  data[myindex + M_YH] = (float) (d->yh);
  data[myindex + M_XTO] = (float) (d->xto);
  data[myindex + M_YTO] = (float) (d->yto);
  data[myindex + M_ATO] = (float) (d->ato);
  data[myindex + M_XHEAD] = (float) (d->xhead);
  data[myindex + M_YHEAD] = (float) (d->yhead);

  data[myindex + M_XSLD] = (float) (d->xsld);
  data[myindex + M_YSLD] = (float) (d->ysld);
  data[myindex + M_XKLD] = (float) (d->xkld);
  data[myindex + M_YKLD] = (float) (d->ykld);
  data[myindex + M_XTHLD] = (float) (d->xthld);
  data[myindex + M_YTHLD] = (float) (d->ythld);
  data[myindex + M_XSRD] = (float) (d->xsrd);
  data[myindex + M_YSRD] = (float) (d->ysrd);
  data[myindex + M_XKRD] = (float) (d->xkrd);
  data[myindex + M_YKRD] = (float) (d->ykrd);
  data[myindex + M_XTHRD] = (float) (d->xthrd);
  data[myindex + M_YTHRD] = (float) (d->ythrd);
  data[myindex + M_XHD] = (float) (d->xhd);
  data[myindex + M_YHD] = (float) (d->yhd);
  data[myindex + M_XTOD] = (float) (d->xtod);
  data[myindex + M_YTOD] = (float) (d->ytod);
  data[myindex + M_XHEADD] = (float) (d->xheadd);
  data[myindex + M_YHEADD] = (float) (d->yheadd);

  data[myindex + M_WSL] = (float) (d->wsl);
  data[myindex + M_WTHL] = (float) (d->wthl);
  data[myindex + M_WSR] = (float) (d->wsr);
  data[myindex + M_WTHR] = (float) (d->wthr);
  data[myindex + M_WTO] = (float) (d->wto);

  data[myindex + M_XCOM] = (float) (d->xcom);
  data[myindex + M_YCOM] = (float) (d->ycom);

  data[myindex + M_XCOMD] = (float) (d->xcomd);
  data[myindex + M_YCOMD] = (float) (d->ycomd);

  data[myindex + M_ENERGY] = (float) (d->energy);
  data[myindex + M_LNE_ENERGY] = (float) (d->lne_energy);

  data[myindex + M_SLKE] = (float) (d->slke);
  data[myindex + M_SLPE] = (float) (d->slpe);
  data[myindex + M_THLKE] = (float) (d->thlke);
  data[myindex + M_THLPE] = (float) (d->thlpe);
  data[myindex + M_SRKE] = (float) (d->srke);
  data[myindex + M_SRPE] = (float) (d->srpe);
  data[myindex + M_THRKE] = (float) (d->thrke);
  data[myindex + M_THRPE] = (float) (d->thrpe);
  data[myindex + M_TOKE] = (float) (d->toke);
  data[myindex + M_TOPE] = (float) (d->tope);
  data[myindex + M_KE] = (float) (d->ke);
  data[myindex + M_PE] = (float) (d->pe);

  myindex += N_CHANNELS;
  n_points++;

  return 0;
}

/*****************************************************************************/

int write_matlab_file( char *filename )
{
  FILE *stream;
  int i;
  static int myindex = 0;

  stream = fopen( filename, "w" );
  if ( stream == NULL )
    {
      fprintf( stderr, "Can't open %s for write.\n", filename );
      exit( -1 );
    }
  for ( i = 0; i < n_points; i++ )
    {
      fprintf( stream, "%g ", data[myindex + M_TIME] );
      // add other stuff you want here
      fprintf( stream, "\n" );
      myindex += N_CHANNELS;
    }
  fclose( stream );

  return 0;
}

/*****************************************************************************/

int write_the_mrdplot_file( Dynamics *s )
{
  MRDPLOT_DATA *d; 

  d = malloc_mrdplot_data( 0, 0 );
  d->filename = generate_file_name();
  d->n_channels = N_CHANNELS;
  d->n_points = n_points;
  d->total_n_numbers = d->n_channels*d->n_points;
  d->frequency = (float) (1/s->time_step);
  d->data = data;
  d->names = names;
  d->units = units;

  write_mrdplot_file( d );

  // write_matlab_file( "dd" );

  return 0;
}

/*****************************************************************************/
