matlab: http://www.mathworks.com/help/toolbox/optim/ug/quadprog.html

x = quadprog(H,f,A,b) minimizes 1/2*x'*H*x + f'*x subject to the
  restrictions A*x ≤ b. A is a matrix of doubles, and b is a vector
  of doubles.

  x = quadprog(H,f,A,b,Aeq,beq) solves the preceding problem subject
  to the additional restrictions Aeq*x = beq. Aeq is a matrix of
  doubles, and beq is a vector of doubles. If no inequalities exist,
  set A = [] and b = [].

  x = quadprog(H,f,A,b,Aeq,beq,lb,ub) solves the preceding problem
  subject to the additional restrictions lb ≤ x ≤ ub. lb and ub are
  vectors of doubles, and the restrictions hold for each x
  component. If no equalities exist, set Aeq = [] and beq = [].

  x = quadprog(H,f,A,b,Aeq,beq,lb,ub,x0) solves the preceding problem
  starting from the vector x0. If no bounds exist, set lb = [] and ub
  = []. Some quadprog algorithms ignore x0. If you do not give x0, all
  components of x0 are set to a point in the interior of the box
  defined by the bounds.

H = eye(10)
f = zeros(10,1)
A = []
b = []

% straight up case

Aeq = [  0 0 0 0 0 0 -1 0 -1 0
         0 0 0 0 0 0 0 -1 0 -1
         0 0 1 0 0 1 -1 -0 -1 -0
        -1 0 1 0 0 0 -1 -0 0 0
         0 0 0 -1 0 1 0 0 0 -1
         0 1 -1 0 0 0 0 0.5 0 0
         0 0 0 0 1 -1 0 0 0 0.5
]
beq = [ -0
-882.9
-0
-0
-0
0
0
]

x =
         0
         0
         0
         0
         0
         0
         0
  441.4500
         0
  441.4500

% legs spaced apart case

Aeq = [  0 0 0 0 0 0 -1 0 -1 0
 0 0 0 0 0 0 0 -1 0 -1
 0 0 1 0 0 1 -0.974695 -0.2 -0.974695 0.2
 -1 0 1 0 0 0 -0.974695 -0.2 0 0
 0 0 0 -1 0 1 0 0 -0.974695 0.2
 0 1 -1 0 0 0 0.497382 0.0510948 0 0
 0 0 0 0 1 -1 0 0 0.477313 -0.148905
]
beq = [ -0
-882.9
-9.59519
-24.4176
14.8224
2.5062
-7.3038
]

x = quadprog(H,f,A,b,Aeq,beq)

x =
  -21.6858
   11.3063
   20.1036
   21.6858
   17.8675
  -29.8291
  -22.5894
  441.1243
   22.5894
  441.7757

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% now don't penalize contact forces and try to make the ankle joints small

H = eye(10)
H(3,3) = 1e4;
H(6,6) = 1e4;
H(7,7) = 1e-6;
H(8,8) = 1e-6;
H(9,9) = 1e-6;
H(10,10) = 1e-6;

x = quadprog(H,f,A,b,Aeq,beq)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

x0 = [ 0
0
0
0
0
0
0
441.5
0
441.5
]

x = quadprog(H,f,A,b,Aeq,beq,[],[],x0)
